#!/usr/bin/env python3
# -*- coding:utf-8 -*-

from __future__ import annotations

import inspect
import sys
from datetime import datetime
from typing import Any, Generic, TypeVar

from typing_extensions import get_args, get_origin

from pendingai.exceptions import ValidationError

Union: Any
if sys.version_info.minor < 10:
    from typing import Union as tUnion  # type: ignore

    Union = tUnion
else:
    from types import UnionType  # type: ignore

    Union = UnionType


__supported_builtins__: tuple[type] = (datetime,)

T = TypeVar("T", bound="Object")


class Object(dict):
    def __init__(self, **fields: Any):
        super().__init__(**fields)
        annotations: dict[str, type] = self.__parsed_annotations__
        fields = self.__defaults__ | fields

        # enforce all required fields defined as resource fields from
        # class attributes to be required in the provided fields if the
        # attribute type hint is non-nullable
        for field_name, field_type in annotations.items():
            if (
                field_name not in fields
                and type(None) not in get_args(field_type)
                and field_type is not None
            ):
                raise TypeError(f"Missing required field '{field_name}'.")

        for field_name, field_type in annotations.items():
            try:
                setattr(self, field_name, self._cast(field_type, fields[field_name]))
            except TypeError as e:
                raise ValidationError(field_name, fields[field_name], field_type) from e

    @classmethod
    def _cast(cls, obj_type: type[Any], obj: Any) -> Any:
        if obj_type == datetime:
            return datetime.fromisoformat(obj)
        if get_origin(obj_type) is Union:
            for union_type in get_args(obj_type):
                try:
                    return union_type(obj)
                except Exception:
                    pass
            raise TypeError("Unable to cast to union type.")
        return obj_type(obj)

    @property
    def __defaults__(self) -> dict[str, Any]:
        return dict(
            [
                a
                for a in inspect.getmembers(
                    self.__class__, lambda x: not inspect.isroutine(x)
                )
                if not (a[0].startswith("__") and a[0].endswith("__"))
            ]
        )

    @property
    def __parsed_annotations__(self) -> dict[str, type]:
        annotations: dict[str, type | str] = (
            self.__dict__.get("__annotations__", None)
            if isinstance(self, type)
            else getattr(self, "__annotations__", {})
        )
        _annotations: dict[str, type] = {}
        for field_name, field_type in annotations.items():
            if isinstance(field_type, str):
                _annotations[field_name] = eval(field_type)
            else:
                _annotations[field_name] = field_type
        return _annotations

    def __getitem__(self, key: str):
        if not hasattr(self, key):
            raise KeyError(key)
        return getattr(self, key)

    def __setitem__(self, key: str, value: Any) -> None:
        if not hasattr(self, key) or key not in self.__parsed_annotations__:
            raise KeyError(key)
        if not isinstance(value, (dt := self.__parsed_annotations__.get(key, None))):  # type: ignore
            raise ValueError(f"'{key}' requires type '{dt.__name__ if dt else dt}'.")
        setattr(self, key, value)


class ListObject(Generic[T], Object):
    object: str = "list"
    data: list[T]
    has_more: bool



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

from __future__ import annotations

import pathlib
import typing

import rich
import rich.progress
import rich.prompt
import typer
from rich.table import Column, Table, box
from typer import Exit, Typer

from pendingai.api_resources.object import ListObject
from pendingai.cli.console import Console
from pendingai.cli.context import PendingAiContext
from pendingai.cli.shared import JsonOption, LimitOption
from pendingai.cli.utils import catch_exception
from pendingai.services.generator.models import Model

cout = Console()
app = Typer(
    name="generator",
    help=(
        "Generative solutions for novel molecule identification and "
        "similarity searches for supplementing synthetic screening of "
        "targeted structures. For more information refer to the "
        "documentation with <pendingai docs>."
    ),
    short_help="Generative solutions for novel molecule identification.",
    no_args_is_help=True,
)


# region command: models -----------------------------------------------


@app.command(
    "models",
    help=(
        "List molecule generator models available for use."
        "\n\nUnique models IDs are required for specifying which model "
        "to use when sampling. If no models are available please "
        "contact support."
    ),
    short_help="List molecule generator models.",
)
@catch_exception()
def models(ctx: PendingAiContext, json: JsonOption = False, limit: LimitOption = 100):
    # collect model resources by enumerating list pointers; for each
    # model we also want to know the status to render
    models: list[Model] = []
    r: ListObject[Model] = ctx.obj["client"].generator.models.list()
    models.extend(r.data)
    while len(models) < limit:
        r = ctx.obj["client"].generator.models.list(created_after=models[-1].id)
        models.extend(r.data)
        if not r.has_more:
            break
    res: list = [
        (model, ctx.obj["client"].generator.models.status(model.id).status)
        for model in models[:limit]
    ]
    if len(res) == 0:
        cout.print("[warn]! No generator models available.")
        raise Exit(1)
    if json:
        cout.print_json(data=res)
    else:
        t = Table("ID", "Name", Column("Version", style="dim"), "Status", box=box.SQUARE)
        for model, status in res:
            t.add_row(
                model.id,
                model.name if model.name else "[dim i]unknown",
                model.version if model.version else "[dim i]unknown",
                status.title(),
            )
        cout.print(t)


# region command: sample -----------------------------------------------


def _generate_sample_output_file() -> pathlib.Path:
    fix: str = "pendingai_generator_sample"
    cwd: pathlib.Path = pathlib.Path.cwd()
    matches: list = sorted([x for x in cwd.iterdir() if x.name.startswith(fix)])
    count: int = (
        int(matches[-1].with_suffix("").name.split("_")[-1]) if len(matches) else 0
    )
    return cwd / f"{fix}_{count + 1:>03d}.smi"


@app.command(
    "sample",
    help=(
        "Sample molecule SMILES from a generator model and output "
        "results to a file. Specify a model id to perform sampling "
        "against a particular molecule generator model."
    ),
    short_help="Sample molecules from a generator model.",
)
@catch_exception()
def sample(
    ctx: PendingAiContext,
    output_file: typing.Annotated[
        pathlib.Path,
        typer.Option(
            "-o",
            "--output-file",
            default_factory=_generate_sample_output_file,
            show_default=False,
            help=(
                "Output SMILES filepath to store sampled molecules. "
                "Defaults to 'pendingai_generator_sample_XXX.smi' in "
                "the current working directory."
            ),
            writable=True,
            dir_okay=False,
            file_okay=True,
            resolve_path=True,
        ),
    ],
    samples: typing.Annotated[
        int,
        typer.Option(
            "-n",
            "--num-samples",
            help="Number of samples to generate. Defaults to 500.",
            show_choices=False,
            show_default=False,
            metavar="INTEGER",
            min=1,
            max=1_000_000,
        ),
    ] = 500,
    append_flag: typing.Annotated[
        bool,
        typer.Option(
            "-a",
            "--append",
            help=(
                "Flag to append generated molecules to an existing "
                "file and skip the prompt for overwriting files."
            ),
            is_flag=True,
        ),
    ] = False,
    model_id: typing.Annotated[
        str | None,
        typer.Option(
            "--model",
            "-m",
            help=(
                "A specified model id used for the sampling process. "
                "Otherwise sample via any available generator model."
            ),
        ),
    ] = None,
):
    # validate output file and append flags to correctly prepare where
    # sampled molecules are written with an output io wrapper flag
    output_file_flag: str = "w"
    prompt: str = f"[warn]? Would you like to overwrite the file: {output_file.name}"
    if output_file.exists() and append_flag:
        cout.print(f"[warn]! Appending results to file: {output_file.name}")
        output_file_flag = "a"
    elif output_file.exists() and not rich.prompt.Confirm.ask(prompt, console=cout):
        cout.print(f"[warn]! See --append for appending to file: {output_file.name}")
        raise typer.Exit(0)
    writer: typing.Any = output_file.open(output_file_flag)

    # build progress bar to track sampling progress, note that file
    # content is being written on each iteration and does not need to
    # wait until the iteration loop is complete
    progress: rich.progress.Progress = rich.progress.Progress(
        rich.progress.SpinnerColumn(finished_text=""),
        *rich.progress.Progress.get_default_columns(),
        rich.progress.TimeElapsedColumn(),
        transient=True,
    )

    # perform sampling until the requested number of samples is finished
    # and uniquely written to file in minibatches
    all_samples: set[str] = set()
    with progress:
        task: rich.progress.TaskID = progress.add_task("Sampling...", total=samples)
        while not progress.finished:
            result: list[str] = (
                ctx.obj["client"].generator.generate.call(id=model_id, n=500).smiles
            )
            sample: set[str] = set(result) - all_samples
            output: list[str] = [x + "\n" for x in sample][: samples - len(all_samples)]
            all_samples = all_samples.union(output)
            writer.writelines(output)
            progress.update(task, completed=len(all_samples))

    cout.print(f"[success]Sampled {len(all_samples)} molecules: {output_file.name}")



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

from __future__ import annotations

from typing import Literal

from requests import Response

from pendingai.api_resources.interfaces import (
    ListResourceInterface,
    RetrieveResourceInterface,
)
from pendingai.api_resources.object import ListObject, Object
from pendingai.api_resources.parser import cast
from pendingai.exceptions import (
    NotFoundError,
    ServiceUnavailableError,
    UnexpectedResponseError,
)


class Model(Object):
    """
    Model object.
    """

    id: str
    """
    Resource id.
    """
    object: str = "model"
    """
    Resource object type.
    """
    name: str | None
    """
    Optional name of the model.
    """
    desc: str | None
    """
    Optional name of the model.
    """
    version: str | None
    """
    Optional version of the model.
    """
    summary: dict
    """
    Additional summary statistics of the model.
    """
    metadata: dict
    """
    Additional metadata describing specific model features.
    """


class ModelStatus(Object):
    """
    Model status object.
    """

    status: str


class ModelInterface(
    ListResourceInterface[Model],
    RetrieveResourceInterface[Model],
):
    """
    Model resource interface; utility methods for model resources.
    """

    def list(
        self,
        *,
        created_before: str | None = None,
        created_after: str | None = None,
        sort: Literal["asc", "desc"] = "desc",
        size: int = 25,
    ) -> ListObject[Model]:  # FIXME: Update api with additional pagination values.
        if sort not in ["asc", "desc"]:
            raise ValueError(f"'sort' must be one of: {['asc', 'desc']}")
        r: Response = self._requestor.request(
            "GET",
            "/generator/v1/models",
            params={
                "pagination-key": created_after,  # FIXME: To be removed.
                "limit": size,  # FIXME: To be removed.
                "created-before": created_before,
                "created-after": created_after,
                "sort": sort,
                "size": size,
            },
        )
        if r.status_code == 200:
            return cast(ListObject[Model], r.json())
        raise UnexpectedResponseError("GET", "list_model")

    def retrieve(self, id: str, *args, **kwargs) -> Model:
        r: Response = self._requestor.request("GET", f"/generator/v1/models/{id}")
        if r.status_code == 200:
            return cast(Model, r.json())
        elif r.status_code == 404:
            raise NotFoundError(id, "Model")
        elif r.status_code == 503:
            raise ServiceUnavailableError
        raise UnexpectedResponseError("GET", "retrieve_model")

    def status(self, id: str) -> ModelStatus:
        r: Response = self._requestor.request("GET", f"/generator/v1/models/{id}/status")
        if r.status_code == 200:
            return cast(ModelStatus, r.json())
        elif r.status_code == 404:
            raise NotFoundError(id, "Model")
        elif r.status_code == 503:
            raise ServiceUnavailableError
        raise UnexpectedResponseError("GET", "retrieve_model")
